﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using InTheHand.Net;
using InTheHand.Net.Bluetooth;
using InTheHand.Net.Sockets;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using NLog;


namespace WebfleetConnect
{
  class BluetoothCntrlr
  {

    private static Logger logger = LogManager.GetCurrentClassLogger();

    BluetoothClient cli;
    Stream peerStream;
    bool isConnected;
    bool isStopserver;
    Dictionary<Guid, BluetoothListener> serverDict = new Dictionary<Guid, BluetoothListener>();
    Dictionary<String, BluetoothClient> clientDict = new Dictionary<String, BluetoothClient>();


    public class DataReceivedEvent
    {
      private byte[] data;
      public DataReceivedEvent(byte[] data) { this.data = data; }
      public byte[] Data { get { return data; } }
    }

    public delegate void DataReceivedEventHandler(object sender, DataReceivedEvent e);
    public event DataReceivedEventHandler DataReceived;
    protected void OnDataReceived(byte[] data)
    {
      if (DataReceived != null)
      {
        DataReceived(this, new DataReceivedEvent(data));
      }
    }


    public event EventHandler ConnectedToServer;
    protected void OnConnectedToServer()
    {
      if (ConnectedToServer != null)
      {
        ConnectedToServer(this, new EventArgs());
      }
    }

    public event EventHandler DisconnectedFromServer;
    protected void OnDisconnectedFromServer()
    {
      if (DisconnectedFromServer != null)
      {
        DisconnectedFromServer(this, new EventArgs());
      }
    }


    public class ClientConnectedEvent
    {
      private String addr;
      private Guid localGuid;
      public ClientConnectedEvent(String addr, Guid localGuid) { this.addr = addr; this.localGuid = localGuid; }
      public String RemoteAddress { get { return this.addr; } }
      public Guid LocalGuid { get { return this.localGuid; } }
    }

    public delegate void ClientConnectedEventHandler(object sender, ClientConnectedEvent e);
    public event ClientConnectedEventHandler ClientConnected;
    protected void OnClientConnected(String addr, Guid localGuid)
    {
      if (ClientConnected != null)
      {
        ClientConnected(this, new ClientConnectedEvent(addr, localGuid));
      }
    }

    public class ClientDisConnectedEvent
    {
      private String addr;
      public ClientDisConnectedEvent(String addr) { this.addr = addr; }
      public String RemoteAddress { get { return this.addr; } }
    }

    public delegate void ClientDisConnectedEventHandler(object sender, ClientDisConnectedEvent e);
    public event ClientDisConnectedEventHandler ClientDisConnected;
    protected void OnClientDisConnected(String addr)
    {
      if (ClientDisConnected != null)
      {
        ClientDisConnected(this, new ClientDisConnectedEvent(addr));
      }
    }


    public class ServerStartedEvent
    {
      private Guid guid;
      public ServerStartedEvent(Guid guid) { this.guid = guid; }
      public Guid ServiceGuid { get { return this.guid; } }
    }

    public delegate void ServerStartedEventHandler(object sender, ServerStartedEvent e);
    public event ServerStartedEventHandler ServerStarted;
    protected void OnServerStarted(Guid guid)
    {
      if (ServerStarted != null)
      {
        ServerStarted(this, new ServerStartedEvent(guid));
      }
    }

    public class ServerStoppedEvent
    {
      private Guid guid;
      public ServerStoppedEvent(Guid guid) { this.guid = guid; }
      public Guid ServiceGuid { get { return this.guid; } }
    }

    public delegate void ServerStoppedEventHandler(object sender, ServerStoppedEvent e);
    public event ServerStoppedEventHandler ServerStopped;
    protected void OnServerStopped(Guid guid)
    {
      if (ServerStopped != null)
      {
        ServerStopped(this, new ServerStoppedEvent(guid));
      }
    }

    public bool IsConectedToServer { get { return this.isConnected; } }


    public void ConnectAsClient(String addr, int channel, string uuid)
    {
      addr = addr.Replace(":", "");
      BluetoothAddress bdAddr = BluetoothAddress.Parse(addr);
      Guid serviceClass;
      BluetoothEndPoint ep;
      if (uuid == null || uuid.Length == 0)
      {
        serviceClass = BluetoothService.SerialPort;
        ep = new BluetoothEndPoint(bdAddr, serviceClass, channel);
      }
      else
      {
        serviceClass = new Guid("{" + uuid + "}");
        ep = new BluetoothEndPoint(bdAddr, serviceClass);
      }

      logger.Info("Connecting as client to [{0}]...", addr);

      this.cli = new BluetoothClient();
      this.cli.Connect(ep);

      logger.Info("Connected.");

      this.peerStream = cli.GetStream();

      this.isConnected = true;
      OnConnectedToServer();
      //ThreadStart ts = new ThreadStart(ReadLoop);
      //Thread thread = new Thread(ts);
      //thread.IsBackground = true;
      //thread.Start();
      ReadLoopFx fx = ReadLoop;
      fx.BeginInvoke(this.cli, null, null);
    }


    public void DisconnectFromServer()
    {
      if (this.isConnected && this.cli != null)
      {
        Disconnect(this.cli);
        this.isConnected = false;
      }
    }

    private void Disconnect(BluetoothClient conn)
    {
      // prevent double clean up
      if (!conn.Connected)
        return;

      logger.Info("Disconnecting...");

      if (conn == this.cli)
      {
        conn.Close();
        OnDisconnectedFromServer();
        isConnected = false;
      }
      else
      {
        String addr = conn.RemoteEndPoint.Address.ToString();
        conn.Close();
        if (this.clientDict.ContainsKey(addr))
          this.clientDict.Remove(addr);
        OnClientDisConnected(addr);
      }
      logger.Info("Disconnected.");
    }

    public void DisconnectClient(String addr)
    {
      if (this.clientDict.ContainsKey(addr))
      {
        Disconnect(this.clientDict[addr]);
      }
    }


    public void SendData(byte[] data)
    {
      if (this.isConnected)
      {
        this.peerStream.Write(data, 0, data.Length);
      }
    }

    public void SendData(byte[] data, String addr)
    {
      if (this.clientDict.ContainsKey(addr))
      {
        this.clientDict[addr].GetStream().Write(data, 0, data.Length);
      }
    }

    private delegate void ReadLoopFx(BluetoothClient conn);
    private void ReadLoop(BluetoothClient conn)
    {
      logger.Info("Getting stream from connection");
      Stream stream = conn.GetStream();
      logger.Info("Got stream. Now entering read loop.");
      while (true)
      {
        try
        {
          byte[] buf = new byte[1000];
          int readLen = stream.Read(buf, 0, buf.Length);
          logger.Info("Read {0} bytes from stream.", readLen);
          if (readLen == 0)
          {
            logger.Info("EOF. will disconnect now");
            break;
          }
          else
          {
            byte[] payload = new byte[readLen];
            Array.Copy(buf, payload, payload.Length);
            OnDataReceived(payload);
          }
        }
        catch (IOException)
        {
          break;
        }
      }
      Disconnect(conn);
    }

    public void StartServer(string uuid, string name, int channel)
    {
      logger.Info("Start server called");
      Guid serviceClass;
      if (uuid == null || uuid.Length == 0)
      {
        serviceClass = BluetoothService.SerialPort;
      }
      else
      {
        serviceClass = new Guid("{" + uuid + "}");
      }

      // already runnning
      if (this.serverDict.ContainsKey(serviceClass))
      {
        logger.Info("Server with UUID {0} alread running", serviceClass.ToString());
        return;
      }

      ServiceRecordBuilder srb = new ServiceRecordBuilder();
      srb.AddServiceClass(BluetoothService.SerialPort);
      srb.AddServiceClass(serviceClass);
      srb.ProtocolType = BluetoothProtocolDescriptorType.Rfcomm;
      if (name != null && name.Length > 0)
      {
        srb.ServiceName = name;
      }

      logger.Info("Creating BluetoothListener");
      var lsnr = new BluetoothListener(serviceClass, srb.ServiceRecord);
      lsnr.LocalEndPoint.Port = channel;
      logger.Info("Listener created. Now starting Listener thread");

      ListenerThreadFx fx = ListenerThread;
      fx.BeginInvoke(lsnr, serviceClass, null, null);
      this.serverDict.Add(serviceClass, lsnr);
      logger.Info("Listener thread started.");
    }

    public void StopServer(Guid serviceClass)
    {
      if (this.serverDict.ContainsKey(serviceClass))
      {
        this.isStopserver = true;
        logger.Info("Stopping server");
        this.serverDict[serviceClass].Stop();
        logger.Info("Server stopped");
      }
    }

    private delegate void ListenerThreadFx(BluetoothListener lsnr, Guid serviceClass);
    private void ListenerThread(BluetoothListener lsnr, Guid serviceClass)
    {
      logger.Info("Starting BluetoothListener");
      try
      {
        lsnr.Start();
      }
      catch (Exception e)
      {
        logger.Error("Error while starting listener: " + e.Message);
        logger.Error(e.StackTrace);
        MessageBox.Show("Error while starting listener: " + e.Message);
        return;
      }
      logger.Info("Listener started");
      OnServerStarted(serviceClass);
      this.isStopserver = false;
      while (true)
      {
        BluetoothClient conn = null;
        try
        {
          logger.Info("Accepting clients...");
          conn = lsnr.AcceptBluetoothClient();
          logger.Info("Client accepted.");
        }
        catch (Exception e)
        {
            // Don't show error when 'Stop Server' was clicked.
            if (!this.isStopserver)
            {
              logger.Error("Error on accepting connections: " + e.Message);
              MessageBox.Show("Error on accepting connections: " + e.Message);
            }
        }
        if (conn == null)
        {
          break;
        }
        this.clientDict.Add(conn.RemoteEndPoint.Address.ToString(), conn);
        OnClientConnected(conn.RemoteEndPoint.Address.ToString(), serviceClass);
        ReadLoopFx fx = ReadLoop;
        fx.BeginInvoke(conn, null, null);
      }
      this.serverDict.Remove(serviceClass);
      OnServerStopped(serviceClass);
      logger.Info("Listener thread ended");
    }
  }

}
