﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using WeifenLuo.WinFormsUI.Docking;

namespace WebfleetConnect
{
  /// <summary>
  /// Dialog to show and change Bluetooth configuration for LINK.connect
  /// It demonstrates the following WF.connect functions:
  /// - configureLocalAuxDevice
  /// - configureRemoteAuxDevice
  /// - getLocalAuxDeviceConfig
  /// - getRemoteAuxDeviceConfig
  /// - removeRemoteAuxDeviceConfig
  /// </summary>
  public partial class ConfigDlg : DockContent, WfcWindow
  {
    Credentials creds;
    HostEntry selectedEndpoint;
    ObjectSelectorDlg objectSelector;


    List<RemoteConfig> configList = new List<RemoteConfig>();
    BindingSource remoteConfigBs = new BindingSource();


    // Table data model for the 'remote config' (BT config related to 3rd party devices).
    class RemoteConfig
    {
      public bool changed = false;  // will not be displayed

      int? configId;
      String deviceId;
      String friendlyName;
      String serviceName;
      String serviceUuid;
      int? channel;
      String pin;
      int? sppBufferSize;
      int? sppFlushTimeout;

      public int? ConfigId { get { return this.configId; } set { this.configId = value; } }
      public String DeviceId { get { return this.deviceId; } set { this.deviceId = value; } }
      public String FriendlyName { get { return this.friendlyName; } set { this.friendlyName = value; } }
      public String ServiceName { get { return this.serviceName; } set { this.serviceName = value; } }
      public String ServiceUuid { get { return this.serviceUuid; } set { this.serviceUuid = value; } }
      public int? Channel { get { return this.channel; } set { this.channel = value; } }
      public String Pin { get { return this.pin; } set { this.pin = value; } }
      public int? SppBufferSize { get { return this.sppBufferSize; } set { this.sppBufferSize = value; } }
      public int? SppFlushTimeout { get { return this.sppFlushTimeout; } set { this.sppFlushTimeout = value; } }

    }

    public ConfigDlg(Settings settings)
    {
      InitializeComponent();
      this.remoteConfigBs.DataSource = this.configList;
      this.remoteConfigDataGridView.DataSource = this.remoteConfigBs;

      objectSelector = new ObjectSelectorDlg(settings);
      objectSelector.TopLevel = false;
      objectSelector.Parent = objectPanel;
      objectSelector.Location = new Point(0, 0);
      objectSelector.Visible = true;
      objectSelector.SelectedObjectChanged += new EventHandler(objectSelector_SelectedObjectChanged);
      this.Enabled = false;
    }

    public void SetCredentials(Credentials creds, HostEntry selectedEndpoint)
    {
      this.creds = creds;
      this.selectedEndpoint = selectedEndpoint;
      this.Enabled = true;
      this.objectSelector.SetCredentials(creds, selectedEndpoint);
    }



    /// <summary>
    /// Retrieve the BT 'local config' via WF.connect and show it in the input fields.
    /// </summary>
    private void LoadLocalConfig()
    {
      var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
      var aParm = MakeAuthParam(); 
      var gParm = new AuxService.GeneralParameters();
      var lcParam = new AuxService.GetLocalAuxDeviceConfigParameter();
      lcParam.objectNo = this.objectSelector.SelectedObject;

      try
      {
        var response = auxClient.getLocalAuxDeviceConfig(aParm, gParm, lcParam);
        if (response.statusCode != 0)
        {
          HandleError(response);
        }
        else
        {
          var config = response.results[0] as AuxService.LocalAuxDeviceConfig;
          this.serviceNameTextBox.Text = config.serviceName;
          this.serviceUuidTextBox.Text = config.serviceUuid;
          this.authenticationCheckBox.Checked = config.authentication;
          this.addGpsCheckBox.Checked = config.addGpsfix;
          this.addTimestampCheckBox.Checked = config.addTimestamp;
          this.addOdometerCheckBox.Checked = config.addOdometer;
        }
      }
      catch (Exception e)
      {
        MessageBox.Show(e.Message, "Error");
      }
    }

    /// <summary>
    /// Retrieve the BT 'remote device config' via WF.connect. The configuration entries are shown
    /// in the remote config list.
    /// </summary>
    private void LoadRemoteConfigs()
    {
      var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
      var aParm = MakeAuthParam();
      var gParm = new AuxService.GeneralParameters();
      var grcParam = new AuxService.GetRemoteAuxDeviceConfigParameter();
      grcParam.objectNo = this.objectSelector.SelectedObject;
      try
      {
        var rresponse = auxClient.getRemoteAuxDeviceConfig(aParm, gParm, grcParam);
        if (rresponse.statusCode != 0)
        {
          HandleError(rresponse);
        }
        else
        {
          configList.Clear();
          foreach (AuxService.RemoteAuxDeviceConfig cfg in rresponse.results)
          {
            RemoteConfig rc = CreateConfigEntry(cfg);
            configList.Add(rc);
          }
          this.remoteConfigBs.ResetBindings(false);
        }
      }
      catch (Exception e)
      {
        MessageBox.Show(e.Message, "Error");
      }

    }

    private static RemoteConfig CreateConfigEntry(AuxService.RemoteAuxDeviceConfig cfg)
    {
      RemoteConfig rc = new RemoteConfig();
      if (cfg.channelSpecified)
        rc.Channel = cfg.channel;
      rc.ConfigId = cfg.configId;
      rc.DeviceId = cfg.deviceId;
      rc.FriendlyName = cfg.friendlyName;
      rc.Pin = cfg.pin;
      rc.ServiceName = cfg.serviceName;
      rc.ServiceUuid = cfg.serviceUuid;
      if (cfg.sppBufferSizeSpecified)
        rc.SppBufferSize = cfg.sppBufferSize;
      if (cfg.sppFlushTimeoutSpecified)
        rc.SppFlushTimeout = cfg.sppFlushTimeout;
      return rc;
    }


    private void HandleError(AuxService.ServiceOpResult response)
    {
      switch (response.statusCode)
      {
        case 0:
          break;
        case 9198:
          // empty result --> OK
          break;
        case 9195:
          String errorMsg = "";
          foreach (var err in response.errors)
          {
            errorMsg += err.objectName + "." + err.fieldName + ": " + err.msg + "\n\r";
          }
          MessageBox.Show(response.statusCode + " " + response.statusMessage + "\n\r" + errorMsg, "Error");
          break;
        default:
          MessageBox.Show(response.statusCode + " " + response.statusMessage, "Error");
          break;
      }
    }


    /// <summary>
    /// Change a single remote device config entry via WF.connect's 'configureRemoteAuxDevice' function.
    /// </summary>
    /// <param name="cfg"></param>
    private void SendDeviceConfig(RemoteConfig cfg)
    {
      var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
      var aParm = MakeAuthParam();
      var gParm = new AuxService.GeneralParameters();
      var cfgParam = new AuxService.RemoteAuxDeviceConfigParameter();
      cfgParam.objectNo = this.objectSelector.SelectedObject;


      if (cfg.ConfigId.HasValue)
        cfgParam.configId = cfg.ConfigId.Value;
      // DataGridView seems to null empty string values. but <null> in WF.c's config API means 'don't change' that value.
      // So we replace it by empty string.
      cfgParam.deviceId = cfg.DeviceId == null ? "" : cfg.DeviceId;
      cfgParam.friendlyName = cfg.FriendlyName;
      cfgParam.pin = cfg.Pin == null ? "" : cfg.Pin;
      cfgParam.serviceName = cfg.ServiceName == null ? "" : cfg.ServiceName;
      cfgParam.serviceUuid = cfg.ServiceUuid == null ? "" : cfg.ServiceUuid;
      cfgParam.sppBufferSize = getManagedParameter(cfg.SppBufferSize);
      cfgParam.sppFlushTimeout = getManagedParameter(cfg.SppFlushTimeout);
      cfgParam.channel = getManagedParameter(cfg.Channel);

      var response = auxClient.configureRemoteAuxDevice(aParm, gParm, cfgParam);
      HandleError(response);
    }

    /// <summary>
    /// Convert the weird Nullable<int> to WF.c's ManagedIntegerParameter.
    /// The ManagedIntegerParameter allows some kind of tri-state:
    /// - null:  don't change that value (not used here since we always change the complete config entry)
    /// - setNull==true: Remove that value (on case of LINK.connect config, a default will apply then)
    /// - specify a value: Use the value .
    /// </summary>
    private AuxService.ManagedIntegerParameter getManagedParameter(int? value)
    {
      var result = new AuxService.ManagedIntegerParameter();
      if (value == null || !value.HasValue)
      {
        result.setNull = true;
      }
      else
      {
        result.value = value.Value;
        result.valueSpecified = true;
      }
      return result;
    }


    /// <summary>
    /// Removes a remote device config entry. It uses WF.connect's 'removeRemoteAuxDeviceConfig' function.
    /// 
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    private void removeConfigButton_Click(object sender, EventArgs e)
    {
      int row = this.remoteConfigDataGridView.CurrentRow.Index;
      if (row >= 0 && row < this.configList.Count)
      {
        RemoteConfig cfg = this.configList[row];

        if (DialogResult.Yes == MessageBox.Show("This will immediately remove config with config ID " + cfg.ConfigId.Value + ".\r\n\n\r Proceed?", "Confirm removal", MessageBoxButtons.YesNo))
        {
          var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
          var aParm = MakeAuthParam();
          var gParm = new AuxService.GeneralParameters();
          var removeParam = new AuxService.RemoveRemoteAuxDeviceConfigParameter();
          removeParam.configId = cfg.ConfigId.Value;
          removeParam.objectNo = this.objectSelector.SelectedObject;
          var response = auxClient.removeRemoteAuxDeviceConfig(aParm, gParm, removeParam);
          HandleError(response);
          if (response.statusCode == 0)
          {
            this.remoteConfigBs.RemoveAt(row);
          }
        }
      }
    }



    private AuxService.AuthenticationParameters MakeAuthParam()
    {
      return new AuxService.AuthenticationParameters
      {
        accountName = creds.account,
        userName = creds.user,
        password = creds.password,
        apiKey = creds.apiKey
      };
    }


    /// <summary>
    /// Send the remote device config for all entries which have been changed by the user. 
    /// Changes will be detected by using the CellValueChanges event.
    /// </summary>
    private void SendRemoteConfigChanges()
    {
      foreach (var cfg in this.configList)
      {
        if (cfg.changed)
        {
          SendDeviceConfig(cfg);
        }
      }
    }

    /// <summary>
    /// Change LINK's local BT config by using WF.connect's configureLocalAuxDevice function.
    /// </summary>
    private void SendLocalConfig()
    {
      var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
      var aParm = MakeAuthParam();
      var gParm = new AuxService.GeneralParameters();
      var cfgParam = new AuxService.LocalAuxDeviceConfigParameter();
      cfgParam.objectNo = this.objectSelector.SelectedObject;
      cfgParam.addGpsfix = this.addGpsCheckBox.Checked;
      cfgParam.addGpsfixSpecified = true;
      cfgParam.addTimestamp = this.addTimestampCheckBox.Checked;
      cfgParam.addTimestampSpecified = true;
      cfgParam.addOdometer = this.addOdometerCheckBox.Checked;
      cfgParam.addOdometerSpecified = true;
      cfgParam.authentication = this.authenticationCheckBox.Checked;
      cfgParam.authenticationSpecified = true;
      cfgParam.serviceName = this.serviceNameTextBox.Text;
      cfgParam.serviceUuid = this.serviceUuidTextBox.Text;
      var response = auxClient.configureLocalAuxDevice(aParm, gParm, cfgParam);
      HandleError(response);
    }

    private void configureLocalButton_Click(object sender, EventArgs e)
    {
      SendLocalConfig();
    }

    /// <summary>
    /// Event handler which is fired when the user selects an object in the embedded ObjectSelectorDlg.
    /// </summary>
    void objectSelector_SelectedObjectChanged(object sender, EventArgs e)
    {
      LoadLocalConfig();
      LoadRemoteConfigs();
    }

    /// <summary>
    /// Mark the remote device config entry as changed.
    /// </summary>
    private void remoteConfigDataGridView_CellValueChanged(object sender, DataGridViewCellEventArgs e)
    {
      if (this.configList.Count > e.RowIndex && e.RowIndex >= 0)
      {
        this.configList[e.RowIndex].changed = true;
      }
    }

    /// <summary>
    /// Click handler for the "Send edit changes" button.
    /// </summary>
    private void sendChangesButton_Click(object sender, EventArgs e)
    {
      SendRemoteConfigChanges();
      LoadRemoteConfigs();
    }

  }


}
