﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using WeifenLuo.WinFormsUI.Docking;

namespace WebfleetConnect
{
  /// <summary>
  /// The dialog for sending aux device data to a 3rd party device. 
  /// It demonstrates following WF.connect functions:
  /// - sendAuxDeviceData
  /// - resetAuxDeviceData
  /// - clearAuxDeviceDataQueue
  /// </summary>
  public partial class DataOutDlg : DockContent, WfcWindow
  {
    Credentials creds;
    HostEntry selectedEndpoint;
    ObjectSelectorDlg objectSelector;
    Settings settings;
    FileInfo fi;


    public DataOutDlg(Settings settings)
    {
      InitializeComponent();
      this.settings = settings;

      objectSelector = new ObjectSelectorDlg(this.settings);
      objectSelector.TopLevel = false;
      objectSelector.Parent = objectPanel;
      objectSelector.Location = new Point(0, 0);
      objectSelector.Visible = true;

      this.Enabled = false;
    }

    public void SetCredentials(Credentials creds, HostEntry selectedEndpoint)
    {
      this.creds = creds;
      this.selectedEndpoint = selectedEndpoint;
      this.Enabled = true;
      this.objectSelector.SetCredentials(creds, selectedEndpoint);
      if (this.settings.lastDeviceAddr != null)
      {
        this.deviceIdTextBox.Text = this.settings.lastDeviceAddr;
      }
    }


    private void selectFileButton_Click(object sender, EventArgs e)
    {
      OpenFileDialog dlg = new OpenFileDialog();
      if (DialogResult.OK == dlg.ShowDialog())
      {
        this.fi = new FileInfo(dlg.FileName);
        if (this.fi.Exists)
        {
          this.fileNameLabel.Text = fi.Name;
          this.sizeLabel.Text = String.Format("{0:###,###,###,###}", this.fi.Length);
        }
      }
    }

    /// <summary>
    /// Sends the content of the text input field to the 3rd party BT device (ASCII encoded).
    /// </summary>
    private void sendTextButton_Click(object sender, EventArgs e)
    {
      System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding();
      byte [] data = enc.GetBytes(this.inputTextBox.Text);
      SendData(data);
    }

    /// <summary>
    /// Send the content of the selected file to the 3rd party BT device.
    /// </summary>
    private void sendFileButton_Click(object sender, EventArgs e)
    {
      try
      {
        if (this.fi != null)
        {
          FileStream fs = new FileStream(this.fi.FullName, FileMode.Open);
          byte[] data = new byte[this.fi.Length];
          int readSize = fs.Read(data, 0, data.Length);
          if (readSize != this.fi.Length)
          {
            MessageBox.Show("Error reading file", "Error");
          }
          else
          {
            SendData(data);
          }
          fs.Close();
        }
      }
      catch (IOException ex)
      {
        MessageBox.Show(ex.Message, "Error");
      }
    }


    /// <summary>
    /// Send the given data to the 3rd party BT device. The device is addressed by the BD_ADDR
    /// entered in the deviceIdTextBox.
    /// </summary>
    private void SendData(byte[] data)
    {
      var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
      var aParm = MakeAuthParam();
      var gParm = new AuxService.GeneralParameters();
      var auxDeviceDataParam = new AuxService.AuxDeviceDataParameter();
      auxDeviceDataParam.objectNo = this.objectSelector.SelectedObject;
      auxDeviceDataParam.deviceId = this.deviceIdTextBox.Text;
      auxDeviceDataParam.data = data;
      var response = auxClient.sendAuxDeviceData(aParm, gParm, auxDeviceDataParam);
      HandleError(response);
      if (response.statusCode == 0)
      {
        MessageBox.Show("Successfully sent " + data.Length + " bytes.", "Data sent");
      }
    }


    private void HandleError(AuxService.ServiceOpResult response)
    {
      switch (response.statusCode)
      {
        case 0:
          break;
        case 9198:
          // empty result --> OK
          break;
        case 9195:
          String errorMsg = "";
          foreach (var err in response.errors)
          {
            errorMsg += err.objectName + "." + err.fieldName + ": " + err.msg + "\n\r";
          }
          MessageBox.Show(response.statusCode + " " + response.statusMessage + "\n\r" + errorMsg, "Error");
          break;
        default:
          MessageBox.Show(response.statusCode + " " + response.statusMessage, "Error");
          break;
      }
    }
    private AuxService.AuthenticationParameters MakeAuthParam()
    {
      return new AuxService.AuthenticationParameters
      {
        accountName = creds.account,
        userName = creds.user,
        password = creds.password,
        apiKey = creds.apiKey
      };
    }

    /// <summary>
    /// Demonstrates the 'resetAuxDeviceData' function.
    /// </summary>
    private void resetButton_Click(object sender, EventArgs e)
    {
      var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
      var aParm = MakeAuthParam();
      var gParm = new AuxService.GeneralParameters();
      var resetParam = new AuxService.ResetAuxDeviceDataParameter();
      resetParam.objectNo = this.objectSelector.SelectedObject;
      var response = auxClient.resetAuxDeviceData(aParm, gParm, resetParam);
      HandleError(response);
    }

    /// <summary>
    /// Demonstrates the 'clearAuxDeviceDataQueue' function.
    /// </summary>
    private void clearButton_Click(object sender, EventArgs e)
    {
      var auxClient = new AuxService.auxDeviceDataClient(this.selectedEndpoint.AuxEndpoint);
      var aParm = MakeAuthParam();
      var gParm = new AuxService.GeneralParameters();
      var clearParam = new AuxService.ClearAuxDeviceDataQueueParameter();
      clearParam.objectNo = this.objectSelector.SelectedObject;
      if (this.deviceIdTextBox.Text.Length > 0)
        clearParam.deviceId = this.deviceIdTextBox.Text;
      var response = auxClient.clearAuxDeviceDataQueue(aParm, gParm, clearParam);
      HandleError(response);
    }

    private void DataOutDlg_FormClosing(object sender, FormClosingEventArgs e)
    {
      this.settings.lastDeviceAddr = this.deviceIdTextBox.Text;
    }

  }
}
