﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Timers;
using WebfleetConnect.QueueService;
using System.Threading;

namespace WebfleetConnect
{
  /// <summary>
  ///  Main window.
  ///  It demonstrates the following tasks:
  ///  - using queue service to check for aux device data messages
  ///  - how to get the payload from the 'aux device data in' messages
  /// </summary>
  public partial class MainForm : Form
  {
    Credentials credentials;
    HostEntry selectedEndpoint;
    Settings settings;

    // Holds the displayed 'aux device data in' messages
    BindingSource auxDataInSource = new BindingSource();
    List<DataInEntry> auxDataInList = new List<DataInEntry>();

    // Holds display models of all queue messages
    BindingSource msgDataSource = new BindingSource();
    List<MsgDataEntry> msgDataList = new List<MsgDataEntry>();

    // Holds the original queue messages 
    List<QueueServiceData> auxDeviceDataList = new List<QueueServiceData>();

    BluetoothDialog btDlg = new BluetoothDialog();
    Thread btWindowThread;

    QueueServicePoller poller;
    const int WAIT_COUNT = 5;

    private delegate void TimerFx(int tick);
    private delegate void MsgDataEventFx(QueueService.QueueServiceData msg);
    private delegate void ErrorEventFx(QueueService.ServiceOpResult response);


    public MainForm()
    {
      InitializeComponent();

      auxDataInSource.DataSource = auxDataInList;
      auxInDataGridView.DataSource = auxDataInSource;
      msgDataSource.DataSource = msgDataList;
      allDataGridView.DataSource = msgDataSource;

      checkProgressBar.Minimum = 0;
      checkProgressBar.Maximum = WAIT_COUNT;
      checkProgressBar.Enabled = false;

      this.settings = ConfigRepo.LoadSetings();
      if (this.settings == null)
      {
        this.settings = new Settings();
      }

      this.btWindowThread = new Thread(new ThreadStart(ShowBTWindow));
    }


    private void MainForm_Shown(object sender, EventArgs e)
    {
      while (true)
      {
        CredentialDialog credentialDlg = new CredentialDialog(this.settings);
        if (DialogResult.OK == credentialDlg.ShowDialog())
        {
          this.credentials = credentialDlg.Credentials;
          this.selectedEndpoint = credentialDlg.SelectedEndpoint;
          if (CheckCredentials())
          {
            InitQueuePoller();
            return;
          }
        }
        else
        {
          Close();
          return;
        }
      }
    }

    private void InitQueuePoller()
    {
      poller = new QueueServicePoller(this.credentials, this.selectedEndpoint);
      this.poller.WaitSeconds = WAIT_COUNT;
      this.poller.TimerTickedEvent += new QueueServicePoller.TimerTickEventHandler(poller_TimerTickedEvent);
      this.poller.MsgReceived += new QueueServicePoller.MsgDataEventHandler(poller_MsgReceived);
      this.poller.ErrorOccured += new QueueServicePoller.QueueErrorEventHandler(poller_ErrorOccured);
      this.poller.QueueChecked += new EventHandler(poller_QueueChecked);
    }


    /// <summary>
    /// Executes a dummy WF.connect actions just to check the credentials
    /// </summary>
    /// <returns></returns>
    private bool CheckCredentials()
    {
      QueueService.messagesClient qs = new QueueService.messagesClient(this.selectedEndpoint.MessagesEndpoint);
      var aParm = new QueueService.AuthenticationParameters
      {
        accountName = this.credentials.account,
        userName = this.credentials.user,
        password = this.credentials.password,
        apiKey = this.credentials.apiKey
      };
      var gParm = new QueueService.GeneralParameters();
      var msgClass = new QueueService.QueueServiceParameter();
      msgClass.filter = QueueService.QueueServiceMessageClassFilter.AUX_DEVICE_DATA;
      var response = qs.popQueueMessagesExtern(aParm, gParm, msgClass);
      HandleError(response);
      return response.statusCode == 0 || response.statusCode == 9198;
      
    }

    void poller_ErrorOccured(object sender, QueueErrorEvent e)
    {
      HandleError(e.ServiceOpResult);
    }

    void poller_MsgReceived(object sender, MsgDataEvent e)
    {
      // Invoke handler method in GUI thread (because it changes GUI)
      Invoke(new MsgDataEventFx(HandleMessage), new object[] { e.Msg });
    }

    void poller_TimerTickedEvent(object sender, TimerTickEvent e)
    {
      if (this.timerCheckBox.Checked && !IsDisposed)
      {
        try
        {
          // Invoke handler method in GUI thread (because it changes GUI)
          Invoke(new TimerFx(TimerWorker), new object[]{e.TicksLeft});
        }
        catch (ObjectDisposedException)
        { }
      }
    }

    void poller_QueueChecked(object sender, EventArgs e)
    {
      Invoke(new MethodInvoker(UpdateLastCheckTime));
    }


    private void TimerWorker(int tick)
    {
      if (tick >= this.checkProgressBar.Minimum && tick < this.checkProgressBar.Maximum)
      {
        this.checkProgressBar.Value = tick;
      }
    }



    
    /// <summary>
    /// Add the queue message to the message DataGridView (lower view). If it's an 'aux device data in' message the incoming data
    /// will also be added to the upper DataGridView which contains all incoming aux data.
    /// </summary>
    /// <param name="msg"></param>
    private void HandleMessage(QueueService.QueueServiceData msg)
    {
      if (msg.auxDeviceData == null)
      {
        return;
      }

      auxDeviceDataList.Add(msg);

      MsgDataEntry msgData = new MsgDataEntry();
      msgData.MsgType = msg.msgType;
      msgData.Text = msg.msgText;
      msgData.Time = msg.msgTime;
      this.msgDataSource.Add(msgData);
      ScrollDown(this.allDataGridView);

      if (msg.auxDeviceData.quotaInfo != null)
      {
        this.quotaMaxLabel.Text = String.Format("{0,15:###,###,###,###}", msg.auxDeviceData.quotaInfo.max);
        this.quotaUsedLabel.Text = String.Format("{0,15:###,###,###,###}", msg.auxDeviceData.quotaInfo.used);
      }
      
      // Special handling for incoming data (show data in list)
      if (msg.msgType == 80000862)
      {
        DataInEntry entry = new DataInEntry();
        entry.Time = msg.msgTime;
        System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding();

        // this is the actual payload (byte array) send from 3rd party BT device
        byte[] payload = msg.auxDeviceData.dataIn.data;

        // decoded as ASCII for displaying purposes
        entry.Data = enc.GetString(payload);
        entry.Device = msg.auxDeviceData.dataIn.deviceId;
        auxDataInSource.Add(entry);
        ScrollDown(this.auxInDataGridView);
      }

    }

    private void ScrollDown(DataGridView grid)
    {
      if (grid.RowCount > 0)
      {
        grid.FirstDisplayedScrollingRowIndex = grid.RowCount - 1;
      }
    }

    /// <summary>
    /// Simple error handler.
    /// </summary>
    /// <param name="response"></param>
    private void HandleError(QueueService.ServiceOpResult response)
    {
      this.timerCheckBox.Checked = false;
      switch (response.statusCode)
      {
        case 9303:
          if (response.errors.Length > 0 && response.errors[0].msg.StartsWith("WFCQ_E0022"))
          {
            if (DialogResult.Yes == MessageBox.Show("Message queue of msg class 15 does not exists. Create it?", "Queue error", MessageBoxButtons.YesNo))
            {
              if (this.poller.CreateQueue())
              {
                MessageBox.Show("Message queue created successfully", "Queue created");
              }
            }
          }
          break;
        case 9198:
          // empty result is OK
          break;
        case 0:
          // empty result is OK
          break;
        default:
          MessageBox.Show(response.statusCode + " " + response.statusMessage, "Error");
          break;
      }
    }



    /// <summary>
    /// Enables or disables the auto check timer.
    /// </summary>
    private void SwitchAutocheck(bool enable)
    {
      if (enable)
      {
        checkProgressBar.Enabled = true;
        checkProgressBar.Value = WAIT_COUNT;
        this.poller.StartPolling();
      }
      else
      {
        this.poller.StopPolling();
        checkProgressBar.Enabled = false;
        checkProgressBar.Value = 0;
      }
    }


    private void SaveState()
    {
      if (this.credentials != null)
      {
        this.settings.account = this.credentials.account;
        this.settings.user = this.credentials.user;
        this.settings.apiKey = this.credentials.apiKey;
        this.settings.server = this.selectedEndpoint.Name;
      }

      ConfigRepo.SaveState(this.settings);
    }

    
    private void checkDataButton_Click(object sender, EventArgs e)
    {
      this.poller.GetQueueData();
    }

    private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
    {
      if (this.poller != null)
        this.poller.StopPolling();

      SaveState();
    }

    private void timerCheckBox_CheckedChanged(object sender, EventArgs e)
    {
      SwitchAutocheck(this.timerCheckBox.Checked);
    }


    private void allDataGridView_CellContentDoubleClick(object sender, DataGridViewCellEventArgs e)
    {
      if (e.RowIndex >= 0)
      {
        ShowMessageDetails(auxDeviceDataList[e.RowIndex]);
      }
    }

    /// <summary>
    /// Open the message detail dialog.
    /// </summary>
    /// <param name="msg"></param>
    private void ShowMessageDetails(QueueServiceData msg)
    {
      AuxDataMessageDlg dlg = new AuxDataMessageDlg();
      dlg.ShowMessage(msg);
      dlg.Dispose();
    }


    private void auxDeviceConfigToolStripMenuItem_Click(object sender, EventArgs e)
    {
      ConfigDlg dlg = new ConfigDlg(this.credentials, this.selectedEndpoint, this.settings);
      dlg.ShowDialog();
      dlg.Dispose();
    }

    private void sendDataToolStripMenuItem_Click(object sender, EventArgs e)
    {
      DataOutDlg dlg = new DataOutDlg(this.credentials, this.selectedEndpoint, this.settings);
      dlg.ShowDialog();
      dlg.Dispose();
    }

    private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
    {
      CopyrightDialog dlg = new CopyrightDialog();
      dlg.ShowDialog();
    }

    private void UpdateLastCheckTime()
    {
      this.lastCheckLabel.Text = DateTime.Now.ToString("G");
    }

    private void bluetoothWindowToolStripMenuItem_Click(object sender, EventArgs e)
    {
      if (this.btWindowThread.ThreadState != ThreadState.Running)
      {
        this.btWindowThread = new Thread(new ThreadStart(ShowBTWindow));
        this.btWindowThread.SetApartmentState(ApartmentState.STA);
        this.btWindowThread.Start();
      }
    }

    private void ShowBTWindow()
    {
      var btDlg = new BluetoothDialog();
      btDlg.ShowDialog();
    }

  }
}
