﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using WeifenLuo.WinFormsUI.Docking;

namespace WebfleetConnect
{
  public partial class MainWindow : Form
  {
    Credentials credentials;
    HostEntry selectedEndpoint;
    Settings settings;

    private WfConnectForm wfConnectForm;
    private BluetoothDialog btForm;
    private DataOutDlg dataOutDlg;
    private ConfigDlg configDlg;

    List<DockContent> windowsList = new List<DockContent>();


    public MainWindow()
    {
      InitializeComponent();

      this.settings = ConfigRepo.LoadSetings();
      if (this.settings == null)
      {
        this.settings = new Settings();
      }

      CreateChildWindows();

      if (this.settings.layoutData != null)
      {
        RestoreLayout(this.settings.layoutData);
      }
      else
      {
        this.btForm.Show(this.dockPanel);
        this.wfConnectForm.Show(this.dockPanel);
      }
    }

    private void CreateChildWindows()
    {
      this.wfConnectForm = new WfConnectForm(this.settings);
      this.btForm = new BluetoothDialog(this.settings);
      this.dataOutDlg = new DataOutDlg(this.settings);
      this.configDlg = new ConfigDlg(this.settings);

      AddChildWindow(this.wfConnectForm);
      AddChildWindow(this.btForm);
      AddChildWindow(this.dataOutDlg);
      AddChildWindow(this.configDlg);
    }

    private void AddChildWindow(DockContent win)
    {
      this.windowsList.Add(win);
      win.DockAreas = DockAreas.Document | DockAreas.Float;
      win.HideOnClose = true;
    }


    private void ShowCredentialDlg()
    {
      while (true)
      {
        CredentialDialog credentialDlg = new CredentialDialog(this.settings);
        if (DialogResult.OK == credentialDlg.ShowDialog())
        {
          this.credentials = credentialDlg.Credentials;
          this.selectedEndpoint = credentialDlg.SelectedEndpoint;
          if (CheckCredentials())
          {
            this.wfConnectForm.SetCredentials(this.credentials, this.selectedEndpoint);
            this.dataOutDlg.SetCredentials(this.credentials, this.selectedEndpoint);
            this.configDlg.SetCredentials(this.credentials, this.selectedEndpoint);
            return;
          }
        }
        else
        {
          return;
        }
      }
    }

    /// <summary>
    /// Simple error handler.
    /// </summary>
    /// <param name="response"></param>
    private void HandleError(ObjectsService.ServiceOpResult response)
    {
      switch (response.statusCode)
      {
        case 9198:
          // empty result is OK
          break;
        case 0:
          // empty result is OK
          break;
        default:
          MessageBox.Show(response.statusCode + " " + response.statusMessage, "Error");
          break;
      }
    }


    /// <summary>
    /// Executes a dummy WF.connect actions just to check the credentials
    /// </summary>
    /// <returns></returns>
    private bool CheckCredentials()
    {
      var objectClient = new ObjectsService.objectsAndPeopleReportingClient(this.selectedEndpoint.ObjectsEndpoint);
      var aParm = new ObjectsService.AuthenticationParameters
      {
        accountName = this.credentials.account,
        userName = this.credentials.user,
        password = this.credentials.password,
        apiKey = this.credentials.apiKey
      };
      var gParm = new ObjectsService.GeneralParameters();
      var objectFilter = new ObjectsService.ObjectFilterParameter();
      var response = objectClient.showObjectReport(aParm, gParm, objectFilter);
      HandleError(response);
      return response.statusCode == 0 || response.statusCode == 9198;
    }


    private void MainWindow_Shown(object sender, EventArgs e)
    {
      ShowCredentialDlg();
    }

    private void wFcLoginToolStripMenuItem_Click_1(object sender, EventArgs e)
    {
      ShowCredentialDlg();
    }

    private void bluetoothToolStripMenuItem_Click_1(object sender, EventArgs e)
    {
      this.btForm.Show(this.dockPanel);
    }

    private void wFcSendAuxDataToolStripMenuItem_Click(object sender, EventArgs e)
    {
      //if (this.dataOutDlg.IsDisposed)
      //{
      //  this.dataOutDlg = new DataOutDlg(this.settings);

      //  if (this.credentials != null)
      //  {
      //    this.dataOutDlg.SetCredentials(this.credentials, this.selectedEndpoint);
      //  }
      //}
      this.dataOutDlg.Show(this.dockPanel);
    }

    private void wFcQueueMessagesToolStripMenuItem_Click(object sender, EventArgs e)
    {
      //if (this.wfConnectForm.IsDisposed)
      //{
      //  this.wfConnectForm = new WfConnectForm(this.settings);
      //  if (this.credentials != null)
      //  {
      //    this.wfConnectForm.SetCredentials(this.credentials, this.selectedEndpoint);
      //  }
      //}
      this.wfConnectForm.Show(this.dockPanel);
    }

    private void wFcConfigurationToolStripMenuItem_Click(object sender, EventArgs e)
    {
      //if (this.configDlg.IsDisposed)
      //{
      //  this.configDlg = new ConfigDlg(this.settings);
      //  if (this.credentials != null)
      //  {
      //    this.configDlg.SetCredentials(this.credentials, this.selectedEndpoint);
      //  }
      //}
      this.configDlg.Show(this.dockPanel);
    }

    private void SetCedentials(WfcWindow win)
    {
      if (this.credentials != null)
      {
        win.SetCredentials(this.credentials, this.selectedEndpoint);
      }
    }


    private void SaveState()
    {
      if (this.credentials != null)
      {
        this.settings.account = this.credentials.account;
        this.settings.user = this.credentials.user;
        this.settings.apiKey = this.credentials.apiKey;
        this.settings.server = this.selectedEndpoint.Name;
      }
      ConfigRepo.SaveState(this.settings);
    }


    private string SaveLayout()
    {
      MemoryStream memStream = new MemoryStream(2000);
      this.dockPanel.SaveAsXml(memStream, Encoding.UTF8, true);
      memStream.Seek(0, SeekOrigin.Begin);
      StreamReader r = new StreamReader(memStream);
      string resultXml = r.ReadToEnd();
      r.Close();
      return resultXml;
    }

    private void RestoreLayout(string layoutXml)
    {
      if (layoutXml != null)
      {
        MemoryStream memStream = new MemoryStream(2000);
        StreamWriter w = new StreamWriter(memStream);
        w.Write(layoutXml);
        w.Flush();
        memStream.Seek(0, SeekOrigin.Begin);
        this.dockPanel.LoadFromXml(memStream, this.DeserializeDockContent, true);
        w.Dispose();
      }
    }

    private IDockContent DeserializeDockContent(string persistString)
    {
      foreach (DockContent win in this.windowsList)
      {
        if (persistString.Equals(win.GetType().ToString()))
        {
          return win;
        }
      }
      return null;
    }

    private void MainWindow_FormClosing(object sender, FormClosingEventArgs e)
    {
      this.settings.layoutData = SaveLayout();
      CloseChildWindows();
      SaveState();
    }

    private void MainWindow_FormClosed(object sender, FormClosedEventArgs e)
    {
    }

    private void CloseChildWindows()
    {
      foreach (DockContent win in this.windowsList)
      {
        win.Close();
      }
    }


  }
}
