﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Timers;
using System.Windows.Forms;
using WebfleetConnect.QueueService;

namespace WebfleetConnect
{
  /// <summary>
  /// Polls the WEBFLEET.connect queue service periodically for new messages.
  /// Various events are fired from this class. E.g. subscribe to 'MsgReceived' to get messages
  /// received from queue service.
  /// Received messages are acknowledged on the WF.c queue. So they are lost if no subsriber is listening while polling.
  /// </summary>
  public class QueueServicePoller
  {
    private Credentials credentials;
    private HostEntry selectedEndpoint;
    private QueueService.QueueServiceMessageClassFilter msgFilter;

    private int waitCounter;
    private int waitSeconds;
    private System.Timers.Timer timer = new System.Timers.Timer();


    public QueueServicePoller(Credentials credentials, HostEntry selectedEndpoint, QueueService.QueueServiceMessageClassFilter msgFilter)
    {
      this.credentials = credentials;
      this.selectedEndpoint = selectedEndpoint;
      this.msgFilter = msgFilter;
      timer.Interval = 1000;
      timer.Elapsed += new ElapsedEventHandler(timer_Elapsed);
      timer.Enabled = false;
    }

    public bool PollingEnabled { get { return this.timer.Enabled; } }
    public int WaitSeconds { get { return this.waitSeconds; } set { this.waitSeconds = value; } }
  
    /// <summary>
    /// Start the perioding polling.
    /// </summary>
    public void StartPolling()
    {
      this.waitCounter = this.waitSeconds;
      this.timer.Enabled = true;
    }

    /// <summary>
    /// Stop the periodic polling.
    /// </summary>
    public void StopPolling()
    {
      this.timer.Enabled = false;
    }

    /// <summary>
    /// Calls WF.connect's popQueueMessagesExtern function to check for new messages. 
    /// </summary>
    public void GetQueueData()
    {
      QueueService.messagesClient qs = new QueueService.messagesClient(this.selectedEndpoint.MessagesEndpoint);
      var aParm = new QueueService.AuthenticationParameters
      {
        accountName = this.credentials.account,
        userName = this.credentials.user,
        password = this.credentials.password,
        apiKey = this.credentials.apiKey
      };
      var gParm = new QueueService.GeneralParameters();
      var msgClass = new QueueService.QueueServiceParameter();
      msgClass.filter = this.msgFilter;

      // pop new messages
      var response = qs.popQueueMessagesExtern(aParm, gParm, msgClass);

      // raise 'QueueChecked' event (can be used by GUI to show that msg check has been done). 
      OnQueueChecked();

      // For every message raise a 'MsgReceived' event
      if (response.statusCode == 0)
      {
        foreach (QueueService.QueueServiceData msg in response.results)
        {
          OnMsgReceived(msg);
        }

        // ACK all received messages to WF.c
        qs.ackQueueMessagesExtern(aParm, gParm, msgClass);
      }
      else if (response.statusCode != 9198)
      {
        // raise 'QueueError' event
        OnQueueError(response);
      }
    }


    /// <summary>
    /// Creates the queue subscription with msg class given in CT.
    /// </summary>
    public bool CreateQueue()
    {
      QueueService.messagesClient qs = new QueueService.messagesClient(this.selectedEndpoint.MessagesEndpoint);
      var aParm = new QueueService.AuthenticationParameters
      {
        accountName = this.credentials.account,
        userName = this.credentials.user,
        password = this.credentials.password,
        apiKey = this.credentials.apiKey
      };
      var gParm = new QueueService.GeneralParameters();
      var msgClass = new QueueService.QueueServiceParameter();
      msgClass.filter = this.msgFilter;
      var response = qs.createQueueExtern(aParm, gParm, msgClass);
      if (response.statusCode != 0)
      {
        OnQueueError(response);
        return false;
      }
      else
      {
        return true;
      }
    }

    void timer_Elapsed(object sender, ElapsedEventArgs e)
    {
      --this.waitCounter;
      OnTimerTicked();
      if (this.waitCounter <= 0)
      {
        this.waitCounter = this.waitSeconds;
        GetQueueData();
      }
    }


    #region Events

    public delegate void TimerTickEventHandler(object sender, TimerTickEvent e);
    public event TimerTickEventHandler TimerTickedEvent;
    protected void OnTimerTicked()
    {
      if (TimerTickedEvent != null)
      {
        TimerTickedEvent(this, new TimerTickEvent(this.waitCounter));
      }
    }
  
    /// <summary>
    /// Fired when a message from the queue message service was received
    /// </summary>
    public event MsgDataEventHandler MsgReceived;
    public delegate void MsgDataEventHandler(object sender, MsgDataEvent e);
    protected void OnMsgReceived(QueueServiceData data)
    {
      if (MsgReceived != null)
      {
        MsgReceived(this, new MsgDataEvent(data));
      }
    }

    public event QueueErrorEventHandler ErrorOccured;
    public delegate void QueueErrorEventHandler(object sender, QueueErrorEvent e);
    protected void OnQueueError(QueueService.ServiceOpResult result)
    {
      if (ErrorOccured != null)
      {
        ErrorOccured(this, new QueueErrorEvent(result));
      }
    }

    public event EventHandler QueueChecked;
    protected void OnQueueChecked()
    {
      if (QueueChecked != null)
      {
        QueueChecked(this, new EventArgs());
      }
    }
    #endregion
  }

#region Event arguments
  public class TimerTickEvent
  {
    private int ticksLeft;
    public TimerTickEvent(int ticksLeft) { this.ticksLeft = ticksLeft; }
    public int TicksLeft { get { return this.ticksLeft; } }
  }

  public class MsgDataEvent
  {
    private QueueServiceData msg;
    public MsgDataEvent(QueueServiceData msg) { this.msg = msg; }
    public QueueServiceData Msg { get { return this.msg; } }
  }

  public class QueueErrorEvent
  {
    private QueueService.ServiceOpResult response;
    public QueueErrorEvent(QueueService.ServiceOpResult response) { this.response = response; }
    public QueueService.ServiceOpResult ServiceOpResult { get { return this.response; } }
  }
#endregion

}
